/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>

#include "bem.h"

/************************************************** show01 *****************/
/* a table of one computed equilibrium */
int show01(){
  int i,ii,iii,m;
  double a,c,d,s,t,p;
  double pgam[MAXCHRG+1];

  d=Cl->ionstr;
  c=sqrt(d);
  a=0.5*(c/(1.0+c) - 0.3*d); /* Davies' equation used here */
  for(i=0;i<=MAXCHRG;i++){
    pgam[i]=a*i*i;
  }

  fprintf(Sfil.fil,"\n=============== Cell %s, Step %d ===============\n"
          ,Cl->id,Cl->step);


  fprintf(Sfil.fil,"\n%16s %7s %8s %8s %8s %8s %8s\n"," ","pa  ","mol(i)/L","mol(e)/L"
        ,"mol  ","%(c) "," mol");

/*                                     H     */
  s=Cl->el[0]->csng;
  p=-log(s)/M_LN10 + pgam[1];
  fprintf(Sfil.fil,"%16s %7.3f %8.4g","H",p,s*1.e6);
  if(Cl->niex>0 && Cl->xc->xcap > 0.){
    s=Cl->xc->xs[0];
    fprintf(Sfil.fil," %8s %8s %8.4g"," "," ",100*s);
  }
  fprintf(Sfil.fil,"\n");

/*                               other elements */
  for(i=2;i<Cl->nelt;i++){
    s=Cl->el[i]->csng;
    c=s;   /* calculate solved total of element */
    for(ii=0;ii<Cl->ncpl;ii++){
      for(iii=0;iii<Cl->cp[ii]->nelt;iii++){
	if(Cl->cp[ii]->el[iii]->num == i){
	  c += Cl->cp[ii]->stcf[iii] * Cl->cp[ii]->ccpl;
	}
      }
    }
    t=c*Cl->theta;
    m=abs(Cl->el[i]->chrg);
    p=-log(s)/M_LN10 + pgam[m];
    fprintf(Sfil.fil,"%16s %7.3f %8.4g %8.4g %8.4g"
           ,Cl->el[i]->name,p,s*1.e6,c*1.e6,t*1.e6);
    if(Cl->el[i]->chrg > 0 && Cl->niex>0 && Cl->xc->xcap > 0.){
      s=Cl->xc->xs[i];
      t=s*Cl->xc->xcap/Cl->el[i]->chrg*Cl->rho;
      fprintf(Sfil.fil," %8.4g   %8.4g",s*100,t*1.e6);
    }
    fprintf(Sfil.fil,"\n");
  }

/*                                     OH    */
  s=Cl->el[1]->csng;
  p=-log(s)/M_LN10 + pgam[1];
  fprintf(Sfil.fil,"%16s %7.3f %8.4g\n","OH",p,s*1.e6);


/*                                     salts                  */
  if(Cl->nslt > 0){
    fprintf(Sfil.fil,"\n%24s %8s %8s %5s %8s\n"
            ," "," ","mmol/kg"," ","  mol ");
    
    for(i=0;i<Cl->nslt;i++){
      c = Cl->sl[i]->cslt;
      t=c*Cl->rho;
      c += Cl->sl[i]->coff;
      a=c*Cl->rho;
      fprintf(Sfil.fil,"%24s %8s %8.4g    %8.4g (%8.4g)\n"
	      ,Cl->sl[i]->name," ",c*1.e3,a*1.e6,t*1.e6);
    }
  }
/*                                         CO2, I, theta, sigma   */
  fprintf(Sfil.fil,"\n");
  for(i=0;i < Cl->ngas; i++){
    p=Cl->cp[Cl->gs[i]->cpnum]->ccpl/Cl->gs[i]->khen;
    fprintf(Sfil.fil,"%6s: %8.4g mbar",Cl->gs[i]->name,p*1.e3);
    ii=Cl->gs[i]->cpnum;                              /*15.6.96*/
    c=Cl->cp[ii]->ccpl * 1.e6 / Cl->gs[i]->khen / RGAS / Cl->temp;
    fprintf(Sfil.fil," (%8.4g mol)\n ",Cl->sigma*c); /*15.6.96*/

  }
  fprintf(Sfil.fil,"I = %g",Cl->ionstr*1.e6);
  fprintf(Sfil.fil,", theta = %g",Cl->theta);
  fprintf(Sfil.fil,", sigma = %g",Cl->sigma);
  fprintf(Sfil.fil,", temp = %g Celsius",Cl->temp-273.15);
  fprintf(Sfil.fil,"\n");


/*                                        comlexes             */
  fprintf(Sfil.fil,"\ncomplexes (mol/L):\n");
  for(i=0;i < Cl->ncpl; i++){
    fprintf(Sfil.fil,"%s %8.4g, ",Cl->cp[i]->name,Cl->cp[i]->ccpl*1.e6);
   }
  fprintf(Sfil.fil,"\n");

  return 0;
}

/*------------------------------------------------ putprp ----------------*/
/* compile put statements into key integers and pointers */
int putprp(){

  int i,k,kk,nwords;
  char rec[MAXINREC], *wrd[MAXWORDS];

  Nputs=0;
  fprintf(Res.fil,"#\n# column definitions:\n# Cell id\n# Step No.\n");

  k=-1;
  rewind(Putsfil);
  while(1){
    if(fgets(rec,MAXINREC,Putsfil)==NULL)break;;
    fprintf(Res.fil,"# %s",rec);
    nwords=words(rec,wrd);
    if(nwords<2){
      bemerr(31,wrd[0],"","",0);
      return -32;
    }
    k++;

    if(strcmp(wrd[1],"pH") == 0){
      Puts[k]=1;
      Cl->putpnt[k]=&Cl->el[0]->pact;
      continue;
    }

    if(strcmp(wrd[1],"ctl") == 0){
      if(nwords<3){
	bemerr(31,"put/ctl","","",0);
	return -32;
      }
      Puts[k]=2;
      kk=-1;
      for(i=0;i<Cl->nelt;i++){
	if(strcmp(wrd[2],Cl->el[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->el[kk]->cttl;
      }
      continue;
    }

    if(strcmp(wrd[1],"sng") == 0){
      if(nwords<3){
	bemerr(31,"put/sng","","",0);
	return -32;
      }
      Puts[k]=3;
      kk=-1;
      for(i=0;i<Cl->nelt;i++){
	if(strcmp(wrd[2],Cl->el[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->el[kk]->csng;
      }
      continue;
    }

    if(strcmp(wrd[1],"xs%") == 0){
      if(nwords<3){
	bemerr(31,"put/xs%","","",0);
	return -32;
      }
      Puts[k]=4;
      kk=-1;
      for(i=0;i<Cl->nelt;i++){
	if(strcmp(wrd[2],Cl->el[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->xc->xs[kk];
      }
      continue;
    }

    if(strcmp(wrd[1],"slt") == 0){
      if(nwords<3){
	bemerr(31,"put/slt","","",0);
	return -32;
      }
      Puts[k]=5;
      kk=-1;
      for(i=0;i<Cl->nslt;i++){
	if(strcmp(wrd[2],Cl->sl[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->sl[kk]->csltt;
      }
      continue;
    }

    if(strcmp(wrd[1],"cpl") == 0){
      if(nwords<3){
	bemerr(31,"put/cpl","","",0);
	return -32;
      }
      Puts[k]=6;
      kk=-1;
      for(i=0;i<Cl->ncpl;i++){
	if(strcmp(wrd[2],Cl->cp[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->cp[kk]->ccpl;
      }
      continue;
    }


    if(strcmp(wrd[1],"p") == 0){
      if(nwords<3){
	bemerr(31,"put/p","","",0);
	return -32;
      }
      Puts[k]=7;
      kk=-1;
      for(i=0;i<Cl->nelt;i++){
	if(strcmp(wrd[2],Cl->el[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=&Cl->el[kk]->pact;
      }
      continue;
    }

    if(strcmp(wrd[1],"dcs") == 0){
      if(nwords<3){
	bemerr(31,"put/dcs","","",0);
	return -32;
      }
      Puts[k]=8;
      kk=-1;
      for(i=0;i<Cl->nelt;i++){
	if(strcmp(wrd[2],Cl->el[i]->name) == 0){
	  kk=i;
	  break;
	}
      }
      if(kk == -1){
	Cl->putpnt[k]=NULL;
      }
      else{
	Cl->putpnt[k]=Cl->el[kk];
      }
      continue;
    }



    bemerr(15,wrd[1],wrd[0],"",0);
    return -32;
  }

  Nputs=k+1;
  fprintf(Res.fil,"# DATA:\n");
  return 0;
}


/*---------------------------------------------- putout ---------------*/


int putout(){
  int i,ii,iii,z;
  double a=0, *p;
  ELT e,*pe;
  static double a0[MAXPUTVAR];

  fprintf(Res.fil,"%10s %10d ",Cl->id,Cl->step);

  for(i=0;i<Nputs;i++){
    if(Cl->putpnt[i]==NULL){
      a=-999.;
    }
    else{ 
      switch(Puts[i]){
      case 1:                     /* pH   */
	p=Cl->putpnt[i];
	a=*p;
 	break;
      case 2:                     /* ctl  */
	p = Cl->putpnt[i];
	a = *p * 1.e6;
	break;
      case 3:                     /* sng  */
	p= Cl->putpnt[i];
	a = *p * 1.e6;
	break;
      case 4:                     /* xs%  */
	p = Cl->putpnt[i];
	a = *p * 1.e2;
	break;
      case 5:                     /* slt  */
	p = Cl->putpnt[i];
	a = *p * 1.e3;
	break;
      case 6:                     /* cpl  */
	p = Cl->putpnt[i];
	a = *p * 1.e6;
	break;
      case 7:                     /* p  */
	p = Cl->putpnt[i];
	a = *p;
	break;
      case 8:                     /* dcs  */
	pe = Cl->putpnt[i];
	e=*pe;
	a=0;
	for(ii=0;ii<Cl->nslt;ii++){
	  for(iii=0;iii<Cl->sl[ii]->nelt;iii++){
	    if(e.num == Cl->sl[ii]->el[iii]->num){
	      a += Cl->sl[ii]->csltt * Cl->sl[ii]->stcf[iii];
	    }
	  }
	}
	if(Cl->niex == 1){
	  z=e.chrg;
	  if(z > 0){
	    a += Cl->xc->xs[e.num]/z *Cl->xc->xcap;
	  }
	}
	if(Cl->step == 0){
	  a0[i]=a;
	  a=0;
	}
	else{
	  a = (a-a0[i])*1.e3;
	}
	break;

       }
    }
    fprintf(Res.fil,"%10g ",a);
  }

  fprintf(Res.fil,"\n");
  return 0;
}

/*------------------------------------------ outint ----------------------*/
/* output of optimization results in internal format */
int outint(int cell, int step){
  int ibuf[3],i;
  double sbuf[MAXSLT];
  char cbuf[MAXSLT];

  ibuf[0]=cell;
  ibuf[1]=step;
  ibuf[2]=Cl->nelt + Cl->niex + Cl->nslt;
  fwrite(ibuf,sizeof(int),3,Ifil.fil);
  fwrite(Cl->nus0.xarg,sizeof(double),ibuf[2],Ifil.fil);

  if(Cl->nslt > 0){
    for(i=0;i<Cl->nslt;i++){
      sbuf[i] = Cl->sl[i]->coff;
    }
    fwrite(sbuf,sizeof(double),Cl->nslt,Ifil.fil);
    for(i=0;i<Cl->nslt;i++){
      cbuf[i] = Cl->sl[i]->state;
    }
    fwrite(cbuf,sizeof(char),Cl->nslt,Ifil.fil);
  }

  return 0;
}
/*------------------------------------------ getint ----------------------*/
/* read output of optimization results in internal format */
int getint(int cell, int step){
  int ibuf[3],i;
  double sbuf[MAXSLT];
  char cbuf[MAXSLT];

  fread(ibuf,sizeof(int),3,Qfil.fil);
  if(ibuf[0] != cell || ibuf[1] != step){
    bemerr(41,Qfil.nam,"","",0);
    return -32;
  }

  fread(Cl->nus0.xarg,sizeof(double),ibuf[2],Qfil.fil);

  if(Cl->nslt > 0){
    fread(sbuf,sizeof(double),Cl->nslt,Qfil.fil);
    for(i=0;i<Cl->nslt;i++){
      Cl->sl[i]->coff = sbuf[i];
    }
    fread(cbuf,sizeof(char),Cl->nslt,Qfil.fil);
    for(i=0;i<Cl->nslt;i++){
      Cl->sl[i]->state = cbuf[i];
    }

  }

  return 0;
}

/* .............................................. cextab ..................*/


int cextab(){
  int i;
  if(Op.s < 1 || Cl->xc->spec == 'd'){
    return 0;
  }

  fprintf(Sfil.fil,"========== Cell %s: Cation exchanger definition ==========\n\n",Cl->id);

  fprintf(Sfil.fil,"%8s%14s%14s\n","","pKsel","xs0");
  for(i=0;i<Cl->nelt;i++){
    if(Cl->el[i]->chrg>0){
      fprintf(Sfil.fil,"%8s  %12.6g  %12.6g\n",
         Cl->el[i]->name,
         Cl->xc->lnksel[i]/(-M_LN10),
         Cl->xc->inpars[i]/Cl->xc->xcap*100);
    }
  }
  fprintf(Sfil.fil,"  CEC  =   %12.6g mmolc/kg\n\n\n",Cl->xc->xcap*1000);
  return 0;
}
